function [sm, vsm] = kalcvs2(data, a, F, b, H, var, filt, vfilt, pred, vpred, smT1, vsmT1, predT1, vpredT1)
%KALCVS The Kalman smoothing
%
%   State space model is defined as follows:
%     z(t+1) = a+F*z(t)+eta(t)     (state or transition equation)
%       y(t) = b+H*z(t)+eps(t)     (observation or measurement equation)
%
%   [sm, vsm] = kalcvs(data, a, F, b, H, var, filt, vfilt, pred, vpred, smT1, vsmT1, predT1, vpredT1)
%     note that T (size of data) will just be a subset of the sample because the problme is written recursively 
%     so one does not necessarily have to start at the end of the sample
%     denote the whole sample as Tall so the smoothed states are z(t|Tall)
%   uses backward recursions to compute the smoothed estimate z(t|Tall) and its covariance matrix, P(t|Tall),
%   where Tall is the number of observations in the complete data set.
%
%   The inputs to the KALCVS function are as follows:
%     data is a NyT matrix containing data (y(1), ... , y(T))'.
%        a is an Nz1 vector for a time-invariant input vector in the transition equation.
%        F is an NzNz matrix for a time-invariant transition matrix in the transition equation.
%        b is an Ny1 vector for a time-invariant input vector in the measurement equation.
%        H is an NyNz matrix for a time-invariant measurement matrix in the measurement equation.
%      var is an (Ny+Nz)(Ny+Nz) covariance matrix for the errors in the transition equation and
%                measurement equation noises, that is, [eta(t)', eps(t)']'.
%     filt is an NzT matrix containing filtered states  (z(1|1), ... , z(T|T))'.
%    vfilt is an NzNzT matrix containing mean square error matrices of filtered state vectors (P(1|1), ... , P(T|T))'.
%     pred is an NzT matrix containing one-step forecasts (z(1|0), ... , z(T|T-1))'.
%    vpred is an NzNzT matrix containing mean square error matrices of predicted state vectors (P(1|0), ... , P(T|T-1))'.
%     smT1 is an  Nz1 vector containing the time T+1 smoothed states sm(T+1) (or  z(T+1|Tall) )
%      vsmT1 is an  NzNz covariance matrix containing the variance of sm(T+1) 
%     predT1 is an  Nz1 vector containing the time T+1 predicted states (or  z(T+1|T) )
%      vpredT1 is an  NzNz covariance matrix containing the variance of predT1 ( P(T+1|T) )
%
%    the problem is written in a recursive way
%    if you want the smoothed shocks for the whole sample just use data for t=1,..,T-1 and the 
%    last period's filtered states (and corresponding variance) as dT1 (vdT1)    
%
%   The KALCVS function returns the following output:
%       sm is an NzT matrix containing smoothed state vectors (z(1|T), ... , z(T|T))'.
%      vsm is an NzNzT matrix containing covariance matrices of smoothed state vectors (P(1|T), ... , P(T|T))'.
%
%   This is a M-file for MATLAB.
%   Copyright 2002-2003 Federal Reserve Bank of Atlanta
%   $Revision: 1.1 $  $Date: 2003/03/31 21:38:05 $
%   Iskander Karibzhanov 3-31-03.
%   Master of Science in Computational Finance
%   Georgia Institute of Technology
%===============================================================================================================
% Revision history:
%
%  03/27/2003  -  algorithm and interface were adapted from SAS/IML KALCVS subroutine for use in MATLAB M file
%
%===============================================================================================================*/

T = size(data,2);
Nz = size(a,1);
Ny = size(b,1);

nin = nargin;
if nin~=14
   error('Ten input arguments required.')
end

nout = nargout;
if nout~=2
   error('Two output arguments required.')
end

% check input matrix dimensions
if size(data,1)~=Ny
   error('data and b must have the same number of rows')
end
if size(a,2)~=1
   error('a must be column vector')
end
if any(size(F)~=[Nz Nz])
   error('F must be square')
end
if size(b,2)~=1
   error('b must be column vector')
end
if any(size(H)~=[Ny Nz])
   error('H must be Ny by Nz matrix')
end
if any(size(var)~=[(Ny+Nz) (Ny+Nz)])
   error('var must be (Ny+Nz) by (Ny+Nz) matrix')
end
if any(size(filt)~=[Nz T])
   error('filt must be Nz by T matrix')
end
if any(size(vfilt)~=[Nz Nz T])
   error('vfilt must be Nz by Nz by T matrix')
end
if any(size(pred)~=[Nz T])
   error('pred must be Nz by T matrix')
end
if any(size(vpred)~=[Nz Nz T])
   error('vpred must be Nz by Nz by T matrix')
end
if any(size(smT1)~=[Nz 1])
   error('dT1 must be column vector of length Nz')
end
if any(size(vsmT1)~=[Nz Nz])
   error('vdT1 must be Nz by Nz matrix')
end
if any(size(predT1)~=[Nz 1])
   error('dT1 must be column vector of length Nz')
end
if any(size(vpredT1)~=[Nz Nz])
   error('vdT1 must be Nz by Nz matrix')
end
% if transition equation has intercept different than zero, demean states
if any(a)
  error('if transition equation has intercept different than zero, use meanz = (eye(Nz)-F)\a demean states')
end




% V(t) and R(t) are variances of eta(t) and eps(t), respectively,
% and G(t) is a covariance of eta(t) and eps(t)
V = var(1:Nz,1:Nz);
R = var(Nz+1:end,Nz+1:end);
G = var(1:Nz,Nz+1:end);

sm = zeros(Nz,T);
vsm = zeros(Nz,Nz,T);

smt1 = smT1;
vsmt1 = vsmT1;

predt1 = predT1;;
vpredt1 = vpredT1;

for t = T:-1:1

  ztt = filt(:,t);
  Ptt = vfilt(:,:,t);
  
  PFP = Ptt'*F'/vpredt1; 
  
  smt = ztt + PFP*(smt1 - predt1);
  vsmt = Ptt + PFP*(vsmt1 - vpredt1)*PFP';
  
  sm(:,t) = smt;
  vsm(:,:,t) = vsmt;

  smt1 = smt;
  vsmt1 = vsmt;

  predt1 = pred(:,t);
  vpredt1 = vpred(:,:,t);

end


